@@ -3,7 +3,7 @@
 module Agda.Compiler.MAlonzo.Compiler where
 
 import Control.Applicative
-import qualified Control.Exception as E
+import Control.Monad ((<=<))
 import Control.Monad.Reader
 import Control.Monad.State
 import Data.Char
@@ -14,13 +14,10 @@ import qualified Language.Haskell.Exts.Extension as HS
 import qualified Language.Haskell.Exts.Parser as HS
 import qualified Language.Haskell.Exts.Syntax as HS
 import System.Cmd
-import System.Directory
-import System.Exit
-import System.IO
-import System.Time
-import System.Process
-import System.FilePath hiding (normalise, (<.>))
+import System.Directory (createDirectoryIfMissing)
+import System.FilePath hiding (normalise)
 
+import Agda.Compiler.CallCompiler
 import Agda.Compiler.MAlonzo.Misc
 import Agda.Compiler.MAlonzo.Pretty
 import Agda.Compiler.MAlonzo.Primitives
@@ -32,24 +29,26 @@ import qualified Agda.Syntax.Concrete.Name as CN
 import Agda.Syntax.Internal
 import Agda.Syntax.Literal
 import Agda.TypeChecking.Monad
+import Agda.TypeChecking.Monad.Builtin
 import Agda.TypeChecking.Monad.Options
 import Agda.TypeChecking.Reduce
-import Agda.TypeChecking.Rules.Builtin.Coinduction
+-- import Agda.TypeChecking.Rules.Builtin.Coinduction
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Substitute
+import Agda.TypeChecking.Telescope
 import Agda.TypeChecking.Level (reallyUnLevelView)
 import Agda.Utils.FileName
 import Agda.Utils.Monad
-import qualified Agda.Utils.IO.Locale as LocIO
 import qualified Agda.Utils.IO.UTF8 as UTF8
 import Agda.Utils.Impossible
+import qualified Agda.Utils.HashMap as HMap
 
 #include "../../undefined.h"
 
 compilerMain :: Interface -> TCM ()
 compilerMain mainI =
   -- Preserve the state (the compiler modifies the state).
-  bracket get put $ \_ -> do
+  localState $ do
 
     -- Compute the output directory.
     opts <- commandLineOptions
@@ -107,9 +106,8 @@ imports = (++) <$> hsImps <*> imps where
 definitions :: Definitions -> TCM [HS.Decl]
 definitions defs = do
   kit <- coinductionKit
-  M.fold (liftM2 (++) . (definition kit <.> instantiateFull))
-         declsForPrim
-         defs
+  HMap.foldr (liftM2 (++) . (definition kit <=< instantiateFull))
+             declsForPrim defs
 
 -- | Note that the INFINITY, SHARP and FLAT builtins are translated as
 -- follows (if a 'CoinductionKit' is given):
@@ -120,17 +118,20 @@ definitions defs = do
 --   sharp :: a -> a
 --   sharp x = x
 --
---   flat :: forall a. () -> forall b. () -> b -> b
---   flat _ _ x = x
+--   flat :: a -> a
+--   flat x = x
 -- @
 
 definition :: Maybe CoinductionKit -> Definition -> TCM [HS.Decl]
 -- ignore irrelevant definitions
-definition kit (Defn Forced     _ _  _ _ _ _) = __IMPOSSIBLE__
-definition kit (Defn NonStrict  _ _  _ _ _ _) = __IMPOSSIBLE__
-definition kit (Defn Irrelevant _ _  _ _ _ _) = return []
-definition kit (Defn Relevant   q ty _ _ compiled d) = do
-  checkTypeOfMain q ty
+{- Andreas, 2012-10-02: Invariant no longer holds
+definition kit (Defn Forced     _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
+definition kit (Defn UnusedArg  _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
+definition kit (Defn NonStrict  _ _  _ _ _ _ _ _) = __IMPOSSIBLE__
+-}
+definition kit (Defn Irrelevant _ _  _ _ _ _ _ _) = return []
+definition kit (Defn _          q ty _ _ _ _ compiled d) = do
+  checkTypeOfMain q ty $ do
   (infodecl q :) <$> case d of
 
     _ | Just (HsDefn ty hs) <- compiledHaskell compiled ->
@@ -168,9 +169,9 @@ definition kit (Defn Relevant   q ty _ _ compiled d) = do
           x    = ihname "x" 0
       return $
         [ HS.TypeSig dummy [flat] $ fakeType $
-            "forall a. () -> forall b. () -> b -> b"
+            "forall a. a -> a"
         , HS.FunBind [HS.Match dummy flat
-                               [HS.PWildCard, HS.PWildCard, HS.PVar x]
+                               [HS.PVar x]
                                Nothing
                                (HS.UnGuardedRhs (HS.Var (HS.UnQual x)))
                                (HS.BDecls [])]
@@ -242,7 +243,7 @@ checkCover q ty n cs = do
 conArityAndPars :: QName -> TCM (Nat, Nat)
 conArityAndPars q = do
   def <- getConstInfo q
-  TelV tel _ <- telViewM $ defType def
+  TelV tel _ <- telView $ defType def
   let Constructor{ conPars = np } = theDef def
       n = genericLength (telToList tel)
   return (n - np, np)
@@ -301,7 +302,7 @@ argpatts ps0 bvs = evalStateT (mapM pat' ps0) bvs
   irr (LitP {})   = return False
   irr (ConP q _ ps) =
     (&&) <$> singleConstructorType q
-         <*> (and <$> mapM irr' ps)
+         <*> (andM $ L.map irr' ps)
 
   -- | Irrelevant patterns are naturally irrefutable.
   irr' (Arg _ Irrelevant _) = return $ True
@@ -319,7 +320,7 @@ clausebody b0 = runReaderT (go b0) 0 where
 --   Types are extracted as @()@.
 --   @DontCare@ outside of irrelevant arguments is extracted as @error@.
 term :: Term -> ReaderT Nat TCM HS.Exp
-term tm0 = case tm0 of
+term tm0 = case ignoreSharing tm0 of
   Var   i as -> do n <- ask; apps (hsVarUQ $ ihname "v" (n - i - 1)) as
   Lam   _ at -> do n <- ask; HS.Lambda dummy [HS.PVar $ ihname "v" n] <$>
                               local (1+) (term $ absBody at)
@@ -335,6 +336,7 @@ term tm0 = case tm0 of
   Sort  _    -> return HS.unit_con
   MetaV _ _  -> mazerror "hit MetaV"
   DontCare _ -> return $ rtmError $ "hit DontCare"
+  Shared{}   -> __IMPOSSIBLE__
   where apps =  foldM (\h a -> HS.App h <$> term' a)
 
 -- | Irrelevant arguments are replaced by Haskells' ().
@@ -444,6 +446,7 @@ writeModule (HS.Module l m ps w ex imp ds) = do
         , "ExistentialQuantification"
         , "ScopedTypeVariables"
         , "NoMonomorphismRestriction"
+        , "Rank2Types"
         ]
 
 rteModule :: HS.Module
@@ -466,11 +469,19 @@ rteModule = ok $ parse $ unlines
   ]
   where
     parse = HS.parseWithMode
-              HS.defaultParseMode{HS.extensions = [HS.ExplicitForall]}
-
+              HS.defaultParseMode{HS.extensions = [explicitForAll]}
     ok (HS.ParseOk d)   = d
     ok HS.ParseFailed{} = __IMPOSSIBLE__
 
+explicitForAll :: HS.Extension
+explicitForAll =
+-- GHC 7.0.1 cannot parse the following CPP conditional
+-- error: missing binary operator before token "("
+#if MIN_VERSION_haskell_src_exts(1,12,0)
+  HS.ExplicitForAll
+#else
+  HS.ExplicitForall
+#endif
 
 compileDir :: TCM FilePath
 compileDir = do
@@ -511,6 +522,7 @@ callGHC i = do
   let overridableArgs =
         [ "-O"
         , "-o", mdir </> show outputName
+        , "-Werror"
         ]
       otherArgs       =
         [ "-i" ++ mdir
@@ -519,7 +531,6 @@ callGHC i = do
         , "--make"
         , "-fwarn-incomplete-patterns"
         , "-fno-warn-overlapping-patterns"
-        , "-Werror"
         ]
       args     = overridableArgs ++ opts ++ otherArgs
       compiler = "ghc"
@@ -527,23 +538,4 @@ callGHC i = do
   -- Note: Some versions of GHC use stderr for progress reports. For
   -- those versions of GHC we don't print any progress information
   -- unless an error is encountered.
-
-  reportSLn "" 1 $ "calling: " ++ L.intercalate " " (compiler : args)
-  (_, _, err, p) <-
-    liftIO $ createProcess (proc compiler args){ std_err = CreatePipe }
-
-  errors <- liftIO $ case err of
-    Nothing  -> __IMPOSSIBLE__
-    Just err -> do
-      -- The handle should be in text mode.
-      hSetBinaryMode err False
-      liftIO $ LocIO.hGetContents err
-
-  exitcode <- liftIO $ do
-    -- Ensure that the output has been read before waiting for the
-    -- process.
-    E.evaluate (length errors)
-    waitForProcess p
-  case exitcode of
-    ExitFailure _ -> typeError (CompilationError errors)
-    _             -> return ()
+  callCompiler compiler args
